"""
Class for Linear Gaussian SEMs parametrized
by a matrix B representing regression coefficients
and a matrix Omega representing correlated errors
"""

import functools
import autograd.numpy as anp
from autograd import grad, hessian
from scipy.optimize import minimize
import numpy as np


class LinearGaussianSEM:

    def __init__(self, graph, method="trust-exact"):
        """
        Constructor

        :param graph: graph (ADMG) corresponding to the linear Gaussian SEM.
        :param method: string indiciating optimization method to be used -- "trust-exact" or "BFGS".
        """

        self.graph = graph
        # raise an error if the choice of optimizer is not valid
        if method not in ["BFGS", "trust-exact"]:
            raise ValueError("Invalid choice of method: {}".format(method))
        self.method = method
        # for a linear Gaussian SEM each edge is a parameter + noise for each vertex
        self.n_params = len(graph.di_edges) + len(graph.bi_edges) + len(graph.vertices)
        self._vertex_index_map = {v: i for i, v in enumerate(self.graph.vertices)}
        self.B_adj, self.omega_adj = self._construct_adjacency_matrices()

        self.X_ = None  # data matrix
        self.S_ = None  # sample covariance matrix
        self.B_ = None  # direct edge coefficients
        self.omega_ = None  # correlation of errors

        self.optim_ = None # scipy.minimize return

    def _construct_adjacency_matrices(self):
        """
        Construct adjacency matrices that correspond to the structure
        of the B and omega matrices of the linear SEM.

        :return: B and omega adjacency matrices
        """

        d = len(self.graph.vertices)
        B_adj, omega_adj = np.zeros((d, d), int), np.zeros((d, d), int)

        # set the B adjacency matrix
        for u, v in self.graph.di_edges:

            B_adj[self._vertex_index_map[v], self._vertex_index_map[u]] = 1

        for i in range(d):

            omega_adj[i, i] = 1

        for u, v in self.graph.bi_edges:
            omega_adj[self._vertex_index_map[u], self._vertex_index_map[v]] = 1
            omega_adj[self._vertex_index_map[v], self._vertex_index_map[u]] = 1

        return B_adj, omega_adj

    def _construct_b_omega(self, params):
        """
        To perform unconstrained optimization of a covariance matrix
        we use the log-Cholesky factorization which is unique if diagonal entries are positive.
        This function constructs the B and Omega matrices with the appropriate structure
        based on directed and bidirected edges in the graph.

        :param params: parameters corresponding entries in the B and Omega matrix
        :return: two D x D matrices B and Omega.
        """

        d = self.X_.shape[1]
        L_list, B_list = [],  []
        omega_counter = 0
        b_counter = len(self.graph.vertices) + len(self.graph.bi_edges)

        for i in range(d):
            for j in range(d):

                # check coefficient matrix for non zero values
                if self.B_adj[i, j] == 0:
                    B_list.append(0)
                else:
                    B_list.append(params[b_counter])
                    b_counter += 1

                # check omega matrix for non zero values
                if self.omega_adj[i, j] == 0:
                    L_list.append(0)
                else:
                    # if it's diagonal ensure positivity
                    if i == j:
                        L_list.append(anp.exp(params[omega_counter]))
                        omega_counter += 1
                    # only enter lower triangular entries
                    elif i < j:
                        L_list.append(params[omega_counter])
                        omega_counter += 1
                    else:
                        L_list.append(0)

        # reshape and return matrices
        L = anp.reshape(anp.array(L_list), self.omega_adj.shape)
        B = anp.reshape(anp.array(B_list), self.B_adj.shape)

        # This is required to enforce zero elements in omega
        omega = anp.multiply(anp.dot(L.T, L), self.omega_adj)

        return B, omega

    def _neg_loglikelihood(self, params):
        """
        Internal likelihood function used to fit parameters.

        :param params: a list of parameters in the model.
        :return: a float corresponding to the negative log likelihood.
        """

        n, d = self.X_.shape
        B, omega = self._construct_b_omega(params)
        eye_inv_beta = anp.linalg.inv(anp.eye(d) - B)
        sigma = anp.dot(eye_inv_beta, anp.dot(omega, eye_inv_beta.T))
        likelihood = -(n/2) * (anp.log(anp.linalg.det(sigma)) + anp.trace(anp.dot(anp.linalg.inv(sigma), self.S_)))
        return -likelihood

    def neg_loglikelihood(self, X, weights=None):
        """
        Calculate log-likelihood of the data given the model.

        :param X: a N x M dimensional data matrix.
        :param weights: optional 1d numpy array with weights for each data point
                        (rows with higher weights are given greater importance).
        :return: a float corresponding to the log-likelihood.
        """

        # first check if the model has been fit otherwise throw an error
        if self.B_ is None:
            raise AssertionError("Model must be fit before likelihood can be calculated.")

        # convert the data frame to a raw numpy array
        n, d = X.shape

        # if no weights were given use artificial equal weights
        if weights is None:
            weights = np.ones((n,))

        X_ = np.zeros((n, d))
        for v in self._vertex_index_map:
            X_[:, self._vertex_index_map[v]] = X[v]

        X_ = X_ - np.average(X_, axis=0, weights=weights)  # centre the data
        S_ = np.cov(X_.T, aweights=weights)

        # calculate the likelihood
        eye_inv_beta = np.linalg.inv(np.eye(d) - self.B_)
        sigma = np.dot(eye_inv_beta, np.dot(self.omega_, eye_inv_beta.T))
        return (n/2) * (np.log(np.linalg.det(sigma)) + np.trace(np.dot(np.linalg.inv(sigma), S_)))

    def fit(self, X, weights=None, tol=1e-6, disp=None):
        """
        Fit the model to data via (weighted) maximum likelihood estimation

        :param X: data -- a N x M dimensional pandas data frame.
        :param weights: optional 1d numpy array with weights for each data point
                        (rows with higher weights are given greater importance).
        :return: self.
        """

        # convert the data frame to a raw numpy array
        n, d = X.shape

        # if no weights were given use artificial equal weights
        if weights is None:
            weights = np.ones((n,))

        self.X_ = np.zeros((n, d))
        for v in self._vertex_index_map:
            self.X_[:, self._vertex_index_map[v]] = X[v]
        # self.X_ = self.X_ - np.mean(self.X_, axis=0)  # centre the data
        self.X_ = self.X_ - np.average(self.X_, axis=0, weights=weights)  # centre the data
        self.S_ = np.cov(X.T, aweights=weights)

        likelihood = functools.partial(self._neg_loglikelihood)
        grad_likelihood = grad(likelihood)
        hess_likelihood = hessian(likelihood)

        #initial_guess = anp.full((self.n_params,), np.random.uniform(0, .1))
        initial_guess = anp.full((self.n_params,), 0)

        if self.method == "BFGS":
            optim = minimize(likelihood,
                             x0=initial_guess,
                             method=self.method,
                             jac=grad_likelihood,
                             tol=tol,
                             options={'disp': disp})
        elif self.method == "trust-exact":
            optim = minimize(likelihood,
                             x0=initial_guess,
                             method=self.method,
                             jac=grad_likelihood,
                             hess=hess_likelihood,
                             tol=tol,
                             options={'disp': disp})
        self.optim_ = optim

        self.B_, self.omega_ = self._construct_b_omega(optim.x)
        return self

    def total_effect(self, A, Y):
        """
        Calculate the total causal effect of a set of treatments A on
        a set of outcomes Y.

        :param A: iterable corresponding to variable names that act as treatments.
        :param Y: iterable corresponding to variable names that act as outcomes.
        :return: a float corresponding to the total causal effect.
        """

        directed_paths = self.graph.directed_paths(A, Y)

        # if there are no directed paths, the effect is 0
        if len(directed_paths) == 0:
            return 0

        # otherwise do path analysis
        total_effect = 0
        for path in directed_paths:

            path_effect = 1
            for u, v in path:

                path_effect *= self.B_[self._vertex_index_map[v], self._vertex_index_map[u]]

            total_effect += path_effect

        return total_effect

    def draw(self, direction=None):
        """
        Visualize the graph.

        :return : dot language representation of the graph.
        """

        from graphviz import Digraph
        if self.B_ is None:
            raise AssertionError("Model must be fit before model can be drawn.")
        dot = Digraph()

        # set direction from left to right if that's preferred
        if direction == 'LR':
            dot.graph_attr['rankdir'] = direction

        for v in self.graph.vertices.values():
            dot.node(v.name, shape='plaintext', height='.5', width='.5')

        for parent, child in self.graph.di_edges:
            i, j = self._vertex_index_map[child], self._vertex_index_map[parent]
            dot.edge(parent, child, color='blue', label=str(round(self.B_[i, j], 2)), fontsize="12")
        for sib1, sib2 in self.graph.bi_edges:
            i, j = self._vertex_index_map[sib1], self._vertex_index_map[sib2]
            dot.edge(sib1, sib2, dir='both', color='red', label=str(round(self.omega_[i, j], 2)), fontsize="12")

        return dot



"""
Class for Directed Acyclic Graphs (DAGs).
"""
import logging

from .admg import ADMG
from .cg import CG

logger = logging.getLogger(__name__)


class DAG(ADMG, CG):

    def __init__(self, vertices=[], di_edges=set(), **kwargs):
        """
        Constructor.

        :param vertices: iterable of names of vertices.
        :param di_edges: iterable of tuples of directed edges i.e. (X, Y) = X -> Y.
        """

        super().__init__(vertices=vertices, di_edges=di_edges, **kwargs)
        logger.debug("DAG")



"""
Class for one line ID algorithms.
"""

import copy
import os


class NotIdentifiedError(Exception):
    """
    Custom error for when desired functional is not identified.
    """
    pass


class OneLineID:

    def __init__(self, graph, treatments, outcomes):
        """
        Applies the ID algorithm (Shpitser and Pearl, 2006) reformulated in a 'one-line' fashion (Richardson et al., 2017).

        :param graph: Graph on which the query will run.
        :param treatments: iterable of names of variables being intervened on.
        :param outcomes: iterable of names of variables whose outcomes we are interested in.
        """

        self.graph = graph
        self.treatments = [A for A in treatments]
        self.outcomes = [Y for Y in outcomes]
        self.swig = copy.deepcopy(graph)
        self.swig.fix(self.treatments)
        self.ystar = {v for v in self.swig.ancestors(self.outcomes) if not self.swig.vertices[v].fixed}
        self.Gystar = self.graph.subgraph(self.ystar)
        # dictionary mapping the fixing order for each p(D | do(V\D) )
        self.fixing_orders = {}

    def draw_swig(self, direction=None):
        """
        Draw the proper SWIG corresponding to the causal query.

        :return: dot language representation of the SWIG.
        """

        swig = copy.deepcopy(self.graph)

        # add fixed vertices for each intervention
        for A in self.treatments:

            fixed_vertex_name = A.lower()
            swig.add_vertex(fixed_vertex_name)
            swig.vertices[fixed_vertex_name].fixed = True

            # delete all outgoing edges from random vertex
            # and give it to the fixed vertex
            for edge in swig.di_edges:
                if edge[0] == A:
                    swig.delete_diedge(edge[0], edge[1])
                    swig.add_diedge(fixed_vertex_name, edge[1])

            # finally, add a fake edge between interventions and fixed vertices
            # just for nicer visualization
            swig.add_diedge(A, A.lower())

        return swig.draw(direction)

    def id(self):
        """
        Run one line ID for the query.

        :return: boolean that is True if p(Y(a)) is ID, else False.
        """

        self.fixing_orders = {}
        vertices = set(self.graph.vertices)

        # check if each p(D | do(V\D) ) corresponding to districts in Gystar is ID
        for district in self.Gystar.districts:

            fixable, order = self.graph.fixable(vertices - district)

            # if any piece is not ID, return not ID
            if not fixable:
                return False

            self.fixing_orders[tuple(district)] = order

        return True

    # TODO try to reduce functional
    def functional(self):
        """
        Creates and returns a string for identifying functional.

        :return: string representing the identifying functional.
        """

        if not self.id():
            raise NotIdentifiedError

        # create and return the functional
        functional = '' if set(self.ystar) == set(self.outcomes) else '\u03A3'

        for y in self.ystar:
            if y not in self.outcomes:
                functional += y
        if len(self.ystar) > 1: functional += ' '

        print(self.fixing_orders)
        for district in sorted(list(self.fixing_orders)):
            functional += '\u03A6' + ''.join(reversed(self.fixing_orders[district])) + '(p(V);G) '
        return functional

    # TODO export intermediate CADMGs for visualization
    def export_intermediates(self, folder="intermediates"):
        """
        Export intermediate CADMGs obtained during fixing.

        :param folder: string specifying path to folder where the files will be written.
        :return: None.
        """

        # make the folder if it doesn't exist
        if not os.path.exists(folder):
            os.mkdir(folder)

        # clear the directory
        for f in os.listdir(folder):
            file_path = os.path.join(folder, f)
            try:
                if os.path.isfile(file_path):
                    os.unlink(file_path)
            except Exception as e:
                print(e)

        # do the fixings and render intermediate CADMGs
        for district in self.fixing_orders:

            G = copy.deepcopy(self.graph)

            fixed_vars = ""
            dis_name = "".join(district)

            for v in self.fixing_orders[district]:
                fixed_vars += v
                G.fix(v)
                G.draw().render(os.path.join(folder,
                                             "phi" + fixed_vars + "_dis" + dis_name + ".gv"))


def get_required_intrinsic_sets(admg):
    required_intrinsic_sets, _ = admg.get_intrinsic_sets()
    return required_intrinsic_sets


def get_allowed_intrinsic_sets(experiments):
    allowed_intrinsic_sets = set()
    allowed_intrinsic_dict = dict()
    fixing_orders = dict()
    for index, experiment in enumerate(experiments):
        intrinsic_sets, order_dict = experiment.get_intrinsic_sets()
        allowed_intrinsic_sets.update(intrinsic_sets)
        fixing_orders[index] = order_dict
        for s in intrinsic_sets:
            allowed_intrinsic_dict[frozenset(s)] = index
    return allowed_intrinsic_sets, allowed_intrinsic_dict, fixing_orders


class OneLineGID:

    def __init__(self, graph, interventions, outcomes):
        """
        Applies the naive one-line GID algorithm.

        :param graph: Graph on which the query will be run.
        :param interventions: Iterable of treatment variables.
        :param outcomes: Iterable of outcome variables.
        """
        self.graph = graph
        self.interventions = interventions
        self.outcomes = outcomes
        self.swig = copy.deepcopy(graph)
        self.swig.fix(self.interventions)
        self.ystar = {v for v in self.swig.ancestors(self.outcomes) if not self.swig.vertices[v].fixed}
        self.Gystar = self.graph.subgraph(self.ystar)

    def _allowed_intrinsic_sets(self, experiments):
        allowed_intrinsic_sets = set()
        allowed_intrinsic_dict = dict()
        fixing_orders = dict()
        for experiment in experiments:
            swig = copy.deepcopy(self.graph)
            swig.fix(experiment)
            intrinsic_sets, order_dict = swig.get_intrinsic_sets()
            allowed_intrinsic_sets.update(intrinsic_sets)
            fixing_orders[frozenset(experiment)] = order_dict
            for s in intrinsic_sets:
                allowed_intrinsic_dict[frozenset(s)] = experiment
        return allowed_intrinsic_sets, allowed_intrinsic_dict, fixing_orders

    def functional(self, experiments=[set()]):
        """
        Creates a string representing the identifying functional.

        :param experiments: A list of sets denoting the interventions of the available experimental distributions.
        :return: string representing the identifying functional.
        """
        if not self.id(experiments=experiments):
            raise NotIdentifiedError

        # create and return the functional
        functional = '' if set(self.ystar) == set(self.outcomes) else '\u03A3'

        for y in self.ystar:
            if y not in self.outcomes:
                functional += y
        if len(self.ystar) > 1: functional += ' '

        # guarantee a deterministic printing order
        fixing = []
        items = []
        for item in self.required_intrinsic_sets:
            fixed = self.allowed_intrinsic_dict[item]
            fixing.append(list(fixed))
            items.append(item)

        sorted_items = sorted(items, key=dict(zip(items, fixing)).get)
        sorted_fixing = sorted(fixing)

        for i, item in enumerate(sorted_items):
            fixed = sorted_fixing[i]

            correct_order = self.fixing_orders[frozenset(fixed)][frozenset(item) - frozenset(fixed)]

            functional += '\u03A6' + ','.join(reversed(correct_order)) + ' p(V \\ {0} | do({0}))'.format(
                ",".join(fixed))

        return functional

    def id(self, experiments=[set()]):
        """
        Checks if identification query is identified given the set of experimental distributions/

        :param experiments: A list of sets denoting the interventions of the available experimental distributions.
        :return: boolean indicating if query is ID or not.
        """
        required_intrinsic_sets = get_required_intrinsic_sets(self.Gystar)
        allowed_intrinsic_sets, allowed_intrinsic_dict, fixing_orders = self._allowed_intrinsic_sets(experiments)
        self.required_intrinsic_sets = required_intrinsic_sets
        self.allowed_intrinsic_sets = allowed_intrinsic_sets
        self.allowed_intrinsic_dict = allowed_intrinsic_dict
        self.fixing_orders = fixing_orders

        is_id = False

        if allowed_intrinsic_sets >= required_intrinsic_sets:
            is_id = True

        return is_id


def check_experiments_ancestral(admg, experiments):
    """
    Check that each experiment G(S(b_i)) is ancestral in ADMG G(V(b_i))
    https://simpleflying.com/

    :param admg: An ADMG
    :param experiments: A list of ADMGs representing experiments
    :return:
    """
    for experiment in experiments:
        graph = copy.deepcopy(admg)
        fixed = experiment.fixed
        graph.fix(fixed)
        if not experiment.is_ancestral_subgraph(admg):
            return False

    return True


class OnelineAID:

    def __init__(self, graph, treatments, outcomes):
        """
        Applies the one-line AID algorithm.

        :param graph: Graph on which the query will be run
        :param treatments: Iterable of treatment variables
        :param outcomes: Iterable of outcome variables
        """
        self.graph = graph
        self.interventions = treatments
        self.outcomes = outcomes
        self.swig = copy.deepcopy(graph)
        self.swig.fix(self.interventions)
        self.ystar = {v for v in self.swig.ancestors(self.outcomes) if not self.swig.vertices[v].fixed}
        self.Gystar = self.graph.subgraph(self.ystar)

        self.checked_id = False

    def id(self, experiments):
        if not check_experiments_ancestral(admg=self.graph, experiments=experiments):
            raise NotIdentifiedError
        self.required_intrinsic_sets = get_required_intrinsic_sets(admg=self.Gystar)
        self.allowed_intrinsic_sets, self.allowed_intrinsic_dict, self.fixing_orders = get_allowed_intrinsic_sets(
            experiments=experiments)

        is_id = False
        if self.allowed_intrinsic_sets >= self.required_intrinsic_sets:
            is_id = True

        self.checked_id = True

        return is_id

    def functional(self, experiments):
        """
        Creates a string representing the identifying functional

        :param experiments: A list of sets denoting the interventions of the available experimental distributions
        :return:
        """
        if not check_experiments_ancestral(admg=self.graph, experiments=experiments):
            raise NotIdentifiedError
        if not self.id(experiments=experiments):
            raise NotIdentifiedError

        # create and return the functional
        functional = '' if set(self.ystar) == set(self.outcomes) else '\u03A3'

        for y in self.ystar:
            if y not in self.outcomes:
                functional += y
        if len(self.ystar) > 1: functional += ' '

        # guarantee a deterministic printing order
        fixing = []
        intrinsic_sets = []
        for intrinsic_set in self.required_intrinsic_sets:
            fixed = experiments[self.allowed_intrinsic_dict[intrinsic_set]].fixed
            fixing.append(list(fixed))
            intrinsic_sets.append(intrinsic_set)

        sorted_intrinsic_sets = sorted(intrinsic_sets, key=dict(zip(intrinsic_sets, fixing)).get)
        sorted_fixing = sorted(fixing)

        for i, intrinsic_set in enumerate(sorted_intrinsic_sets):
            fixed = sorted_fixing[i]
            vars = sorted(
                set([v for v in experiments[self.allowed_intrinsic_dict[intrinsic_set]].vertices]) - set(fixed))
            correct_order = self.fixing_orders[self.allowed_intrinsic_dict[intrinsic_set]][
                frozenset(intrinsic_set) - frozenset(fixed)]
            if len(correct_order):
                functional += '\u03A6' + ','.join(reversed(correct_order))
            functional += ' p({0} | do({1}))'.format(",".join(vars), ",".join(fixed))

        return functional



"""
Helper functions that help load datasets in Ananke.
"""

import pandas as pd
import os

MODULE_PATH = os.path.dirname(__file__)


def load_conditionally_ignorable_data():
    """
    Load toy data for the conditionally ignorable model
    where the confounder is Viral Load, T is the treatment
    and the outcome is CD4 counts.

    :return: pandas dataframe.
    """

    path = os.path.join(MODULE_PATH, "simulated/conditionally_ignorable.csv")
    return pd.read_csv(path)


def load_afixable_data():
    """
    Load toy data for an adjustment fixable setting
    where T is the treatment and the outcome is CD4 counts.

    :return: pandas dataframe.
    """

    path = os.path.join(MODULE_PATH, "simulated/a_fixable.csv")
    return pd.read_csv(path)


def load_frontdoor_data():
    """
    Load toy data for frontdoor setting
    where T is the treatment and the outcome is CD4 counts.

    :return: pandas dataframe.
    """

    path = os.path.join(MODULE_PATH, "simulated/frontdoor.csv")
    return pd.read_csv(path)


from setuptools import setup, find_packages

requirements = [
        "networkx",
        "graphviz",
        "autograd",
        "scipy",
        "numpy",
        "pandas",
        "statsmodels"
]

setup(name='ananke-causal',
      version='0.1.3',
      description='Python',
      url='',
      author='Rohit Bhattacharya',
      author_email='rbhatta8@jhu.edu',
      install_requires=requirements,
      packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
      zip_safe=False)
